from functools import wraps
import inspect
import typing


def _find_type_origin(type_hint):
    if isinstance(type_hint, typing._SpecialForm):
        # case of typing.Any, typing.ClassVar, typing.Final, typing.Literal,
        # typing.NoReturn, typing.Optional, or typing.Union without parameters
        return

    actual_type = typing.get_origin(type_hint) or type_hint  # requires Python 3.8
    if isinstance(actual_type, typing._SpecialForm):
        # case of typing.Union[…] or typing.ClassVar[…] or …
        for origins in map(_find_type_origin, typing.get_args(type_hint)):
            yield from origins
    else:
        yield actual_type


def _check_types(parameters, hints):
    for name, value in parameters.items():
        type_hint = hints.get(name, typing.Any)
        actual_types = tuple(_find_type_origin(type_hint))
        if actual_types and not isinstance(value, actual_types):
            raise TypeError(
                f"Expected type '{type_hint}' for argument '{name}'"
                f" but received type '{type(value)}' instead"
            )


def enforce_types(callable):
    def decorate(func):
        hints = typing.get_type_hints(func)
        signature = inspect.signature(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            parameters = dict(zip(signature.parameters, args))
            parameters.update(kwargs)
            _check_types(parameters, hints)

            return func(*args, **kwargs)

        return wrapper

    if inspect.isclass(callable):
        callable.__init__ = decorate(callable.__init__)
        return callable

    return decorate(callable)


def enforce_strict_types(callable):
    def decorate(func):
        hints = typing.get_type_hints(func)
        signature = inspect.signature(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            bound = signature.bind(*args, **kwargs)
            bound.apply_defaults()
            parameters = dict(zip(signature.parameters, bound.args))
            parameters.update(bound.kwargs)
            _check_types(parameters, hints)

            return func(*args, **kwargs)

        return wrapper

    if inspect.isclass(callable):
        callable.__init__ = decorate(callable.__init__)
        return callable

    return decorate(callable)


def accepts_m(*types) -> typing.Callable:
    """Check parameter types for a class-method.

    Returns:
        Callable: Wrapped method that is now typesafe on its inputs.
    """
    return accepts(object, *types)


def accepts(*types) -> typing.Callable:
    """Check parameter types for a function.

    Returns:
        Callable: Wrapped function that is now typesafe on its inputs.
    """

    def check_accepts(f):
        #  assert len(types) == f.func_code.co_argcount
        assert (
            len(types) == f.__code__.co_argcount
        ), f"{f.__code__.co_argcount = } != {len(types) = }"

        def new_f(*args, **kwds):
            for a, t in zip(args, types):
                assert isinstance(a, t), "arg %r does not match %s" % (a, t)
            return f(*args, **kwds)

        # new_f.func_name = f.func_name
        new_f.__name__ = f.__name__
        return new_f

    return check_accepts


def returns(rtype: typing.Type) -> typing.Callable:
    """Check return types of a function.

    Args:
        rtype (Type): Type of the return value.

    Returns:
        Callable: Wrapped function which return value confirm to
        the type-declaration.
    """

    def check_returns(f):
        def new_f(*args, **kwds):
            result = f(*args, **kwds)
            assert isinstance(
                result, rtype
            ), f'return value {result} does not match {rtype} in function "{f.__name__}"'
            return result

        # new_f.func_name = f.func_name
        new_f.__name__ = f.__name__
        return new_f

    return check_returns
